#++ ---------------- RatSol.mu ---------------------

Author: Winfried Fakler,  University of Karlsruhe
Date Created:  8 August 1996
Last Update : 13 August 1996


                **** Beta-Version ****


Description:
This file contains functions for computing polynomial and rational solutions
of linear ordinary differential equations over the field of rational functions.

Functions: 

 - used Parameter:
    ++ Ly, y(x), R:DOM_EXPR
    ++ y, x:DOM_IDENT
    ++ n, m:DOM_INT
    ++ A, a:DOM_LIST
    ++ P, Q:DOM_POLY
    ++ Ly   = ordinary linear differential equation over the rational functions
    ++ y(x) = the operator (function) of Ly  
    ++ y    = name of the operator of Ly
    ++ x    = dependant variable of the operator y
    ++ n    = order of Ly
    ++ m    = exponent of the symmetric power of Ly
    ++ A    = vectorised form Ly (i.e. all coefficients of Ly)
    ++ R    = Ansatz for rational solutions with a bounded but indeterminate
    ++        numerator and determined denominator
    ++ a    = list of the indeterminates of the numerator of R
    ++ P, Q = polynomials

 - ode::rationalSolutions(Ly, y(x))
    ++ returns a fundamental system of rational solutions of an ordinary
    ++ linear differential equation Ly over the rational functions.
 - ode::ratsols(Ly, y, x, n)
    ++ returns a fundamental system of rational solutions of an ordinary
    ++ linear differential equation Ly over the rational functions.
    ++ NOTE: Ly should be normalized (i.e. denom(Ly)=1).
 - ode::rationalvariants(Ly, y(x), m)
    ++ returns a fundamental system of rational solutions of the m-th
    ++ symmetric power of an ordinary linear differential equation Ly
    ++ over the rational functions.
 - ode::ratvars(Ly, y, x, n, m)
    ++ returns a fundamental system of rational solutions of the m-th
    ++ symmetric power of an ordinary linear differential equation Ly
    ++ over the rational functions.
 - ode::ratAnsatz(A, x, n, < "Indets" >)
    ++ returns the Ansatz for computing rational solutions of A.
    ++ When the option "Indets" is given, the indeterminates of the Ansatz
    ++ are returned additionally.
    ++ NOTE: A should be normalized (i.e. all denominators must be one).
 - ode::denomAnsatz(A, x, n < "mu" >)
    ++ returns the Ansatz of the numerator for rational solutions of A.
    ++ When the option "mu" is given, the sum of the exponents 
    ++ are returned additionally.
    ++ NOTE: A should be normalized (i.e. all denominators must be one).
 - ode::polyBound(A, x, n)
    ++ determines an upper bound for polynomial solutions of A.
    ++ NOTE: A should be normalized (i.e. all denominators must be one).
 - ode::ratBasis(A, x, R, a)
    ++ determines a fundamental system of rational solutions of A.
    ++ NOTE: A should be normalized (i.e. all denominators must be one).
 - ode::nu(Q, P, < Rem >)
    ++ computes the order of Q at P, which is the largest integer m
    ++ such that P^m divides Q.
    ++ When the option Rem is given, Q/P^m is returned.


See: e.g. M. Bronstein (1992). Integration and Differential Equations in
     Computer Algebra.  Programmirovanie 18, Nr. 5, pages 9-10,
     http://www.inf.ethz.ch/personal/bronstein/

Acknowledgement: 
This implementation is based on the ode::ratsols implementation
from Paul Zimmermann.

Examples:
>> Ly := y(x)*6 + x*diff(y(x), x)*(-2) + diff(y(x), x, x)*(-x^2 + 1):
>> ode::rationalSolutions(Ly, y(x));
 
                                  2
                                {x  - 1/3}
 
>> ode::rationalvariants(Ly, y(x), 2);
 
                           {      2            }
                           {   2 x     4       }
                           { - ---- + x  + 1/9 }
                           {    3              }
++#

ode::rationalvariants := proc(eq, z, m)
  begin
    eq := ode::isLODE(eq, z, "Lode");
    case nops(eq) 
      of 1 do return(error("not an ordinary linear differential equation ". 
                            "over the rational functions"));
      of 4 do if eq[4] < 1 then
                 return(error("only possible for positive order equations"))
              end_if; 
      otherwise
        if domtype(m) <> DOM_INT or m < 1
          then return(error("only defined for positive integers"))
        end_if
    end_case;
    eq[1] := numer(normal(expr(eq[1])));
    ode::ratvars(eq, m)
  end_proc:

ode::ratvars := proc(eq, y, x, n, m)
  local A, Q, R, ord, a, i, mu, degR;
  begin
    A   := [poly(coeff(eq, diff(y(x), x$i), 1), [x]) $hold(i)=0..n ];
    ode::symmetricPower; # to load ode::symPower #
    eq := ode::symPower(eq, y, x, n, m);
    eq := numer(normal(expr(eq)));
    Q  := ode::denomAnsatz(A, x, n, "mu");
    mu := Q[2]; Q := Q[1];
    indets(eq, RatExpr) minus indets(eq);
    ord := max(-infinity, op(map(%, nops)))-1;
    Q   := Q^(ord-n); 
    A   := [poly(coeff(eq, diff(y(x), x$i), 1), [x]) $hold(i)=0..ord ];
    degR := mu + ode::polyBound(A, x, ord);
    R := 0; 
    for i from 1 to degR+1 do 
      a[i] := genident("a"); R := R+a[i]*x^(i-1) 
    end_for;
    ode::ratBasis(A, x, R/Q, [a[i] $ hold(i)=1..degR+1])
  end_proc:      

ode::rationalSolutions := proc(eq, z)
  begin
    eq := ode::isLODE(eq, z, "Lode");
    case nops(eq) 
      of 1 do return(error("not an ordinary linear differential equation ". 
                            "over the rational functions"));
      of 4 do if eq[4] < 1 then
                 return(error("only defined for positive order equations"))
              end_if;
    end_case;
    eq[1] := numer(normal(expr(eq[1])));
   ode::ratsols(eq)
  end_proc:  

ode::ratsols := proc(eq, y, x, n)
  local i;
  begin
    [poly(coeff(eq, diff(y(x), x$i), 1), [x]) $hold(i)=0..n ];
    ode::ratBasis(%, x, ode::ratAnsatz(%, x, n, "Indets"))
  end_proc:

ode::ratAnsatz := proc(A, x, n, o)  #o: Option "Indets" #
  local k, l, m, Q, mu, i, J, j, v, vmin, P, q, z, d, dp, a, degR, R;
  begin
    #A := map(A, fun(poly(expr(args(1)), [x])) );#
    l := factor(A[n+1]); m := nops(l) div 2; mu := 0;
    for i from 1 to m do
      Q[i] := l[2*i]; 
      J[i] := {}; 
      vmin := infinity;
      for k from 0 to n do
         v := ode::nu(A[k+1], Q[i]) - k;
         if v<vmin then J[i] := {k}; vmin := v
           elif v=vmin then J[i] := J[i] union {k}
         end_if;
      end_for;
      dp := diff(Q[i],x); # from Eckhard.Pfluegel@imag.fr #
      P[i] := 0; 
      for j in J[i] do
         # multiplication by dp^j suggested by Eckhard.Pfluegel@imag.fr #
         P[i] := P[i] + multcoeffs(
                   dp^j*ode::nu(A[j+1],Q[i],Rem),_mult(z-k$hold(k)=0..j-1) )
      end_for;
      P[i] := divide(P[i],Q[i],Rem);
      P[i] := resultant(P[i],Q[i],x);
      # select negative roots #
      d[i]:=select(sharelib::iroots(P[i],FALSE),_less,0);
      if d[i]={} then d[i] := 0 else d[i]:= -min(op(d[i])) end_if;
      mu:=mu+d[i]*degree(Q[i]);
    end_for;
    J := {}; 
    vmin := -infinity; 
    for k from 0 to n do
      v := degree(A[k+1])-k;
      if v>vmin then J := {k}; vmin := v 
        elif v=vmin then J := J union {k} end_if
    end_for;
    P:=0;  
    for j in J do 
      P := P + _mult((-1)^j,lcoeff(A[j+1]), (if iszero(j) then 1 else
                                             z + k $hold(k)=0..j-1 end_if) )  
    end_for;
    q:=sharelib::iroots(P,FALSE);
    if q={} then q := 0 else q := min(op(q)) end_if;
    degR := mu + max(n-1,-q);
    R:=0; 
    for i from 1 to degR+1 do 
      a[i] := genident("a"); R := R+a[i]*x^(i-1) 
    end_for;
    Q := expr(_mult(Q[i]^d[i]$hold(i)=1..m));
    if o="Indets" then R/Q, [a[i] $hold(i)=1..degR+1]
      else R/Q end_if
  end_proc:


ode::denomAnsatz := proc(A, x, n, o) # Option: "mu", the sum of Exponents #
  local k, l, m, Q, mu, i, J, j, v, vmin, P, z, d, dp;
  begin
    l := factor(A[n+1]); m := nops(l) div 2; mu := 0;
    for i from 1 to m do
      Q[i] := l[2*i]; 
      J[i] := {}; 
      vmin := infinity;
      for k from 0 to n do
         v := ode::nu(A[k+1], Q[i]) - k;
         if v<vmin then J[i] := {k}; vmin := v
           elif v=vmin then J[i] := J[i] union {k}
         end_if;
      end_for;
      dp := diff(Q[i],x); # from Eckhard.Pfluegel@imag.fr #
      P[i] := 0; 
      for j in J[i] do
         # multiplication by dp^j suggested by Eckhard.Pfluegel@imag.fr #
         P[i] := P[i] + multcoeffs(
                   dp^j*ode::nu(A[j+1],Q[i],Rem),_mult(z-k$hold(k)=0..j-1) )
      end_for;
      P[i] := divide(P[i],Q[i],Rem);
      P[i] := resultant(P[i],Q[i],x);
      # select negative roots #
      d[i]:=select(sharelib::iroots(P[i],FALSE),_less,0);
      if d[i]={} then d[i] := 0 else d[i] := -min(op(d[i])) end_if;
      mu := mu + d[i]*degree(Q[i]);
    end_for;
    expr(_mult(Q[i]^d[i]$hold(i)=1..m));
    if o="mu" then %, mu else % end_if
  end_proc:

ode::polyBound := proc(A, x, n)
  local J, j, k, v, vmin, P, q, z; 
  begin 
    #A := map(A, fun(poly(expr(args(1)), [x])) ); #
    J := {}; 
    vmin := -infinity; 
    for k from 0 to n do
      v:=degree(A[k+1])-k;
      if v > vmin then J:={k}; vmin:=v 
        elif v = vmin then J := J union {k} end_if
    end_for;
    userinfo(10, "Indexmenge: ", J);
    P := 0; 
    for j in J do 
      P := P+_mult(lcoeff(A[j+1]), (if iszero(j) then 1 else
                                     z - k $hold(k)=0..j-1 end_if) ) 
    end_for;
    userinfo(10, "determinierende Gleichung: ", expand(P));
    q := sharelib::iroots(P, FALSE);
    max(n-1, op(q));
  end_proc:

ode::ratBasis := proc(A, x, R, a)
  local Q, i, sys, unk, e, j, eq;
  begin
    Q[1]:=R; 
    for i from 2 to nops(A) do 
      Q[i] := normal(diff(Q[i-1],x)) 
    end_for;
    sys := numer( _plus(expr(A[i])*Q[i] $hold(i)=1..nops(A)) );
    sys := {coeff(sys,[x])}; 
    unk := {a[i] $hold(i)=1..nops(a)};
    # the following lines avoid an expensive call to linsolve #
    j:=null();
    repeat
      e:=FALSE; for eq in sys do
         i := indets(eq) intersect unk;
         if nops(i)=1 then
            i := op(i);
            if degree(eq,[i]) = 1 then
               sys := sys minus {eq}; unk := unk minus {i};
               eq  := i=-coeff(eq,[i],0)/coeff(eq,[i],1);
               j   := j,eq; sys := subs(sys,eq); 
               e := TRUE; break;
            end_if
         end_if
      end_for;
    until e = FALSE end_repeat;
    sys := sys minus {0};
    if sys <> {} then
      sys := linsolve(sys, unk); 
      if sys=null() or sys=FAIL then return([]) end_if; 
      j := j,op(sys)
    end_if;
    unk := [op(unk)];
    { ( coeff(subs(R,j), [unk[i]], 1); 
       if iszero(%) then null() else normal(%) end_if )
      $ hold(i)=1..nops(unk) }
  end_proc:

# returns the largest integer m such that P^m divides Q
   or Q/P^m is o=Rem #
ode::nu := proc(Q,P,o) 
  local m, R;
  begin
    if iszero(Q) then m := infinity
      else 
        m := 0; 
        repeat
          R:=divide(Q,P);
          if iszero(R[2]) then m := m+1; Q := R[1] else break end_if
        until FALSE end_repeat
    end_if;
    if o=hold(Rem) then Q else m end_if
  end_proc:
    

